"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
const os = require("os");
const logs_1 = require("./logs");
const nanoid_1 = require("nanoid");
const jaeger_client_1 = require("jaeger-client");
const opentracing_1 = require("opentracing");
const zlib = require("zlib");
// import { createTerminus } from '@godaddy/terminus';
const fastify_1 = require("fastify");
const fastify_cors_1 = require("fastify-cors");
const fastify_compress_1 = require("fastify-compress");
/**
 * Clase para crear un transport, remplazando un broker.
 *
 * @export
 * @class HttpTransport
 * @extends {EventEmitter}
 */
class HttpTransport {
    /**
     * Creates an instance of HttpTransport.
     * @param {string} topic
     * @memberof HttpTransport
     */
    constructor(topic, version, appname, // Nombre de la aplicacion
    serviceName, // Nombre del servicio
    service) {
        /**
         * Puerto en el que inicia el servidor http
         *
         * @private
         * @type {number}
         * @memberof HttpTransport
         */
        this.port = process.env.HTTP_PORT
            ? parseInt(process.env.HTTP_PORT, 0)
            : 3000;
        /**
         * Manejador de logs
         *
         * @private
         * @type {*}
         * @memberof HttpTransport
         */
        this.logs = logs_1.default;
        /**
         * Topico del servicio
         *
         * @private
         * @type {string}
         * @memberof HttpTransport
         */
        this.topicName = "";
        /**
         * Instancia del servicio
         *
         * @private
         * @type {Microservice}
         * @memberof HttpTransport
         */
        this.service = null;
        this.topic = topic;
        this.hostname = os.hostname();
        this.appname = appname;
        this.version = version;
        this.serviceName = serviceName;
        this.service = service; // Instancia del servicio
        const JAEGER_AGENT_HOST = process.env.JAEGER_AGENT_HOST || "jaeger-agent.logging";
        this.server = fastify_1.default({
            logger: false ? process.env.NODE_ENV === "PRODUCTION" : true,
        });
        // Registrar Cors
        this.server.register(fastify_cors_1.default);
        // Comprimir
        this.server.register(fastify_compress_1.default, { global: true });
        const config = {
            serviceName: `${this.appname}_${this.version}_${this.serviceName}`,
            reporter: {
                agentHost: JAEGER_AGENT_HOST,
            },
            sampler: {
                param: 1,
                type: "const",
            },
        };
        const options = {
            tags: {
                "http-enabled": true,
            },
        };
        // Iniciar el traecing
        this.tracer = jaeger_client_1.initTracer(config, options);
    }
    /**
     * Metodo para crear los datos necesarios de la metadata.
     *
     * @private
     * @returns {Metadata}
     * @memberof HttpTransport
     */
    createMetaData(operationID, topicName, topic, transactionID, hostname, asynchronous, id, method, deviceType, filters, span) {
        return {
            uuid: id,
            id_operacion: operationID,
            id_transaction: transactionID,
            intents: 1,
            callback: topic,
            callbacks: [{ method, name: topicName }],
            owner: topic,
            uowner: topic,
            uworker: `${topic}_${operationID}`,
            worker: topic,
            asynchronous,
            mtype: "input",
            bifurcacion: false,
            time: "",
            from: hostname,
            method,
            deviceType,
            span,
            filters,
            http: {
                code: 200,
                headers: {},
                message: "",
            },
        };
    }
    /**
     * Metodo para ejecutar una peticion.
     *
     * @private
     * @param {DataRequest} request
     * @returns {Promise<DataRequest>}
     * @memberof HttpTransport
     */
    process(request) {
        return new Promise((rr, ff) => {
            this.service.message(request, rr); // Ejecutar el servicio y enviar el callback de la promesa
        });
    }
    /**
     * Metodo para preparar los datos requeridos de la peticion
     *
     * @private
     * @param {string} method
     * @param {*} data
     * @param {*} filters
     * @param {string} uuid
     * @param {*} headers
     * @returns {DataRequest}
     * @memberof HttpTransport
     */
    prepare(method, data, filters, uuid, headers, span, operationID) {
        // const message = Buffer.from(JSON.stringify({}),).toString('base64')
        const transactionID = nanoid_1.nanoid();
        // Retornar los datos necesarios.
        return {
            metadata: this.createMetaData(operationID, this.topicName, this.topicName, transactionID, this.hostname, false, uuid, method, "PC", filters, span),
            data,
            headers,
        };
    }
    /**
     * Handler para las peticiones desde el cliente.
     *
     * @private
     * @param {Request} req
     * @param {ResponseToolkit} resp
     * @memberof HttpTransport
     */
    handler(req, resp) {
        return __awaiter(this, void 0, void 0, function* () {
            // Tiempo inicial de la operacion
            const start = new Date().getTime();
            const params = req.params;
            // Id unico de la operacion
            const operationID = new Date().getTime() / 1000000 + "";
            // Id de la operacion
            const id = params.id || null;
            // Metodo
            const method = req.method.toUpperCase();
            // Filters
            const filters = req.query;
            // Data
            const data = req.body || {};
            // Headers
            const headers = req.headers;
            // Header Cors
            const headersCors = {
                "Access-Control-Allow-Origin": "*",
                "Access-Control-Allow-Methods": "*",
                "Access-Control-Allow-Headers": "*",
            };
            // Generar el span
            const span = this.tracer.startSpan("Endpoint (Internal Server)");
            // Asignar un Tags
            span.setTag(opentracing_1.Tags.COMPONENT, "HTTP-SERVER");
            span.log({
                headers,
                filters,
                method,
                operationID,
            });
            // Crear la consulta
            const reqOperation = this.prepare(method, data, filters, id, headers, span, operationID);
            try {
                const respOperation = yield this.process(reqOperation);
                const r = respOperation.response;
                let dataGzip = null;
                respOperation.response.meta = Object.assign(respOperation.response.meta, {
                    time_elapsed: new Date().getTime() - start,
                });
                span.log({
                    response: respOperation.response,
                });
                // Validar si se incluye el soporte de GZIP
                if (req.headers["accept-encoding"] &&
                    req.headers["accept-encoding"].includes("gzip")) {
                    // Aplicar gzip
                    try {
                        dataGzip = yield this.gzip(respOperation.response);
                    }
                    catch (error) {
                        this.logs.error("No fue posible aplicar Gzip a los datos: ", error);
                    }
                }
                span.finish();
                if (respOperation.metadata.hasOwnProperty("http")) {
                    this.logs.info(respOperation.metadata.http);
                    if (respOperation.metadata.http.hasOwnProperty("code")) {
                        if (typeof respOperation.metadata.http.code !== "number") {
                            if (r.data.response.errorCode ||
                                r.data.response.errorCode < 0) {
                                respOperation.metadata.http.code = 500;
                            }
                            else {
                                respOperation.metadata.http.code = 200;
                            }
                        }
                    }
                    else {
                        if (r.data.response.errorCode ||
                            r.data.response.errorCode < 0) {
                            respOperation.metadata.http.code = 500;
                        }
                        else {
                            respOperation.metadata.http.code = 200;
                        }
                    }
                    // Respuesta
                    const rp = resp
                        // .response(await this.gzip(respOperation.response))
                        .code(respOperation.metadata.http.code)
                        .send(respOperation.response);
                    // Validar que existan cabeceras
                    if (respOperation.metadata.http.hasOwnProperty("headers")) {
                        for (const key of Object.keys(respOperation.metadata.http.headers)) {
                            rp.header(key, respOperation.metadata.http.headers[key]);
                        }
                    }
                    // Agregar la cabecera de cencoding en caso de aplicar gzip
                    if (dataGzip) {
                        rp.header("Content-Encoding", "gzip");
                        rp.header("Content-Type", "application/json");
                    }
                    // Forzar cabeceras CORS
                    for (const key of Object.keys(headersCors)) {
                        rp.header(key, headersCors[key]);
                    }
                    // Retornar la respuesta
                    return rp;
                }
                else {
                    let rp = null;
                    // Validar si ocurrio un error
                    if (r.data.response.errorCode ||
                        r.data.response.errorCode < 0) {
                        rp = resp
                            .code(500)
                            .send(dataGzip ? dataGzip : respOperation.response);
                    }
                    else {
                        rp = resp.send(dataGzip ? dataGzip : respOperation.response);
                    }
                    // Agregar la cabecera de cencoding en caso de aplicar gzip
                    if (dataGzip) {
                        rp.header("Content-Encoding", "gzip");
                        rp.header("Content-Type", "application/json");
                    }
                    for (const key of Object.keys(headersCors)) {
                        rp.header(key, headersCors[key]);
                    }
                    // Retornar la respuesta
                    return rp;
                }
            }
            catch (error) {
                this.logs.error(error);
                span.setTag(opentracing_1.Tags.ERROR, error);
                span.finish();
                const rp = resp.code(500).send({
                    meta: {
                        status: "ERROR",
                        id_transaction: operationID,
                        time_elapsed: new Date().getTime() - start,
                    },
                    data: {
                        response: {
                            errorCode: -9,
                            userMessage: "-35",
                        },
                    },
                });
                // Forzar cabeceras CORS
                for (const key of Object.keys(headersCors)) {
                    rp.header(key, headersCors[key]);
                }
                return rp;
            }
        });
    }
    gzip(data) {
        return new Promise((rr, ff) => {
            zlib.gzip(JSON.stringify(data), (err, buffer) => {
                if (err) {
                    ff(err);
                }
                else {
                    rr(buffer);
                }
            });
        });
    }
    /**
     *  Metodo que inicia el servidor HTTP.
     *  @public
     */
    start() {
        return __awaiter(this, void 0, void 0, function* () {
            // Escuchar los eventos
            const routes = [
                "/{id}",
                `/${this.appname}/api/${this.version}/${this.serviceName}`,
                `/${this.appname}/api/${this.version}/${this.serviceName}/{id}`,
                `/api/${this.version}/${this.serviceName}`,
                `/api/${this.version}/${this.serviceName}/{id}`,
            ];
            routes.forEach((route) => {
                this.server.route({
                    method: [
                        "DELETE",
                        "GET",
                        "HEAD",
                        "PATCH",
                        "POST",
                        "PUT",
                        "OPTIONS",
                    ],
                    path: route,
                    handler: this.handler.bind(this),
                });
            });
            this.server.route({
                method: [
                    "DELETE",
                    "GET",
                    "HEAD",
                    "PATCH",
                    "POST",
                    "PUT",
                    "OPTIONS",
                ],
                path: "/",
                handler: (req, resp) => {
                    return "ok";
                },
            });
            const startServer = () => __awaiter(this, void 0, void 0, function* () {
                try {
                    yield this.server.listen(this.port, "0.0.0.0", (err, address) => {
                        if (err) {
                            this.logs.error(err);
                            process.exit(1);
                        }
                        this.logs.info(`server listening on ${address}`);
                    });
                }
                catch (err) {
                    this.logs.error(err);
                    process.exit(1);
                }
            });
            // Mandar llamar
            startServer();
        });
    }
}
exports.default = HttpTransport;
